module TT_Find_Curves
  
  
  def self.run
    model = Sketchup.active_model
    
    start_time = Time.now
    Sketchup.status_text = 'Finding curves...'
    
    curves = self.find_curves( model.active_entities )
    
    # Output performance data
    puts '=== Find Curves ==='
    puts 'Method: TT_Find_Curves'
    puts "Found #{curves.size} curves in #{Time.now-start_time}s"
    
    # Visualize results
    materials = (0...50).map { |i|
      m=model.materials.add
      m.color = Sketchup::Color.new( 64+rand(128), 64+rand(128), 64+rand(128) )
      m
    }
    model.start_operation('Visualize Results', true)
    for curve in curves
      current_material = materials[ rand(10) ]
      for edge in curve
        edge.material = current_material
      end
    end
    model.commit_operation
    
    nil
  end # def
  
  
  # Find continous sets of edges not part of faces. The result is an array of
  # more arrays. Each sub-array contains a set of sorted edges.
  #
  # Sub-arrays may contains arrays of only one edge.
  #
  # @param [Array<Sketchup::Entity>] entities
  #
  # @return [Array<Array<Sketchup::Edge>>] An +Array+ of Arrays.
  # @since 2.5.0
  def self.find_curves( entities )
    source_edges = entities.select { |e|
      e.is_a?( Sketchup::Edge ) && e.faces.empty?
    }
    curves = []
    cache = source_edges.to_a
    until cache.empty?     
      curve = []
      stack = [ cache.first ]
      until stack.empty?
        edge = stack.shift
        curve << edge
        # Find connected edges
        valid_vertices = edge.vertices.select { |v|
          v.edges.size == 2
        }
        edges = valid_vertices.map { |v| v.edges }
        edges.flatten!
        edges.uniq!
        edges -= curve
        stack += edges
      end # until stack.empty?
      curves << self.sort( curve )
      cache -= curve
    end # until cache.empty?
    curves
  end
  
  
  # Sorts the given set of edges from start to end. If the edges form a loop
  # an arbitrary start is picked.
  #
  # @todo Comment source
  #
  # @param [Array<Sketchup::Edge>] edges
  #
  # @return [Array<Sketchup::Edge>] Sorted set of edges.
  # @since 2.5.0
  def self.sort( edges )
    forks = edges.any? { |e|
      connected = e.start.edges + e.end.edges
      ( ( connected & edges ) - [e] ).size > 2
    }
    return nil if forks
    ends = edges.select { |e|
      connected = e.start.edges + e.end.edges
      ( ( connected & edges ) - [e] ).size == 1
    }
    return nil if ends.size > 2
    sorted = []
    edge = (ends.empty?) ? edges.first : ends.first
    until edge.nil?
      sorted << edge
      connected = edge.start.edges + edge.end.edges
      edge = ( ( connected & edges ) - sorted ).first
    end
    sorted
  end

end # module